/**********************************************************************
 * greeting.c                                              October 1999
 * Horms                                             horms@valinux.com.
 * VA Linux System                                 http://valinux.com./
 *
 * Protocol independent greeting
 *
 * Released under the terms of the GNU GPL
 *
 **********************************************************************/

#include "greeting.h"

/**********************************************************************
 * greeting
 * Send a greeting to the user
 * Pre: fd: file descriptor to write to
 * Return 0 on success
 *        -1 on error
 **********************************************************************/

int greeting(const int fd, protocol_t *protocol){
  char *message;

  if((message=greeting_str(message, protocol))==NULL){
    syslog(LOG_WARNING, "greeting: greeting_str");
    return(-1);
  }

  if(protocol->write(fd, NULL, protocol->type[PROTOCOL_OK], message)<0){
    syslog(LOG_WARNING, "in_greeting: protocol->write");
    return(-1);
  }

  free(message);
  
  return(0);
}


/**********************************************************************
 * greeting_str
 * Produce greeting string
 * Pre: message: unallocated ponter to put greeting string in
 * Return message string on success
 *        NULL on error
 **********************************************************************/

char * greeting_str(char *message, protocol_t *protocol){
  extern struct utsname *system_uname;

  if((message=cat_str(
    message,
    3,
    protocol->greeting_string,
    " ",
    system_uname->nodename
  ))==NULL){
    syslog(LOG_WARNING, "greeting_str: cat_str");
    return(NULL);
  }

  return(message);
}
