/**********************************************************************
 * perdition.c                                           September 1999
 * Horms                                             horms@valinux.com.
 * VA Linux System                                 http://valinux.com./
 *
 * Perdition, POP3 and IMAP4 proxy daemon
 *
 * Released under the terms of the GNU GPL
 **********************************************************************/

#include "perdition.h"

/*Use uname information here and there to idinify this system*/
struct utsname *system_uname;

/**********************************************************************
 * Muriel the main funtion
 **********************************************************************/

int main (int argc, char **argv){
  struct passwd *pw;
  int client_in;
  int client_out;
  int server;
  unsigned char *buffer;
  options_t opt;
  server_port_t *server_port;
  char *servername;
  char *port;
  token_t *tag;
  protocol_t *protocol;
  int status;

  extern struct utsname *system_uname;

  tag=NULL;

  /*Parse options*/
  opt=options(argc, argv);

  /*Open syslog*/
  openlog(LOG_IDENT, LOG_OPTION, LOG_FACILITY);

  /*Set signal handlers*/
  /*signal(SIGHUP,    (void(*)(int))daemon_exit_cleanly);*/
  signal(SIGINT,    (void(*)(int))daemon_exit_cleanly);
  signal(SIGQUIT,   (void(*)(int))daemon_exit_cleanly);
  signal(SIGILL,    (void(*)(int))daemon_exit_cleanly);
  signal(SIGTRAP,   (void(*)(int))daemon_exit_cleanly);
  signal(SIGIOT,    (void(*)(int))daemon_exit_cleanly);
  signal(SIGBUS,    (void(*)(int))daemon_exit_cleanly);
  signal(SIGFPE,    (void(*)(int))daemon_exit_cleanly);
  signal(SIGUSR1,   (void(*)(int))daemon_exit_cleanly);
  signal(SIGSEGV,   (void(*)(int))daemon_exit_cleanly);
  signal(SIGUSR2,   (void(*)(int))daemon_exit_cleanly);
  signal(SIGPIPE,   (void(*)(int))daemon_exit_cleanly);
  signal(SIGALRM,   (void(*)(int))daemon_exit_cleanly);
  signal(SIGTERM,   (void(*)(int))daemon_exit_cleanly);
  signal(SIGURG,    (void(*)(int))daemon_exit_cleanly);
  signal(SIGXCPU,   (void(*)(int))daemon_exit_cleanly);
  signal(SIGXFSZ,   (void(*)(int))daemon_exit_cleanly);
  signal(SIGVTALRM, (void(*)(int))daemon_exit_cleanly);
  signal(SIGPROF,   (void(*)(int))daemon_exit_cleanly);
  signal(SIGWINCH,  (void(*)(int))daemon_exit_cleanly);
  signal(SIGIO,     (void(*)(int))daemon_exit_cleanly);

  /*Close file descriptors and detactch process from shell as necessary*/
  if(opt.inetd_mode){
    daemon_inetd_process();
  }
  else{
    daemon_process();
  }

  /*Seed the uname structure*/
  if((system_uname=(struct utsname *)malloc(sizeof(struct utsname)))==NULL){
    syslog(LOG_WARNING, "main: malloc");
    daemon_exit_cleanly(-1);
  }
  if(uname(system_uname)){
    syslog(LOG_WARNING, "main: uname");
    daemon_exit_cleanly(-1);
  }

  /*Set up protocol structure*/
  if((protocol=protocol_initialise(opt.protocol, protocol))==NULL){
    syslog(LOG_WARNING, "main: protocol_initialise");
    daemon_exit_cleanly(-1);
  }

  /*Set port now the protocol structure is accessable*/
  if((opt.port=(*(protocol->port))(opt.port))==NULL){
    syslog(LOG_WARNING, "main: protocol->port");
    daemon_exit_cleanly(-1);
  }

  /*Tell the world we are here */
  syslog( LOG_INFO, "Started");
  if(log_options(opt)){
    syslog(LOG_WARNING, "main: log_options");
    daemon_exit_cleanly(-1);
  }

  /*Open incoming socket at required*/
  if(opt.inetd_mode){
    client_in=0;
    client_out=1;
  }
  else{
    if((client_in=tcp_socket_server_connect(opt.port))<0){
      syslog(LOG_WARNING, "main: tcp_socket_server_connect");
      daemon_exit_cleanly(-1);
    }
    client_out=client_in;
  }

  if(opt.authenticate_in){
    if(geteuid()){
      syslog(
        LOG_INFO, 
        "Warning: not invoked as root, local authentication may fail"
      );
    }
  }
  else{
    /*Become someone else*/
    /*NB: We do this later if we are going to authenticate locally*/
    if(geteuid()){
      syslog(LOG_INFO, "Warning: not invoked as root, not setting uid and gid");
    }
    else if(daemon_setid(opt.username, opt.group)){
      syslog(LOG_WARNING, "Error setting group and userid, Exiting");
      daemon_exit_cleanly(-1);
    }
  }

  /*Speak to our client*/
  if(greeting(client_out, protocol)){
    syslog(LOG_WARNING, "main: greeting");
    daemon_exit_cleanly(-1);
  }

  /* Authenticate the user*/
  for(;;){
    /*Read the USER and PASS lines from the client */
    if((pw=(*(protocol->in_get_pw))(client_in, client_out, &tag))==NULL){
      syslog(LOG_WARNING, "main: protocol->in_get_pw");
      daemon_exit_cleanly(-1);
    }

    /*Log the session*/
    syslog(LOG_INFO, "Connect: user=%s passwd=%s", pw->pw_name, pw->pw_passwd);

    /*Read the server from the map*/
    if((server_port=gdbm_getserver(opt, pw->pw_name))!=NULL){
      if((port=server_port_get_port(server_port))==NULL){
        port=opt.port;
      }
      servername=server_port_get_servername(server_port);
    }
    else{
      servername=NULL;
      port=NULL;
    }

    /*Log the session*/
    syslog(
      LOG_INFO, 
      "Connect: user=%s passwd=%s server=%s port=%s", 
      pw->pw_name,
      pw->pw_passwd,
      servername,
      port
    );

    if(opt.authenticate_in){
      /*Authenticate user locally - Not really necessary*/
      if((status=protocol->in_authenticate(pw, client_out, tag))==0){
        syslog(LOG_INFO, "Local authentication failure");
        continue;
      }
      else if(status<0){
        syslog(LOG_WARNING, "main: pop3_in_authenticate");
        break;
      }
    }

    /*Try again ig we didn't get anything useful*/
    if(server_port==NULL){
      sleep(PERDITION_ERR_SLEEP);
      if((*(protocol->write))(
        client_out, 
	tag,
	protocol->type[PROTOCOL_ERR], 
	"Could not determine server"
      )<0){
        syslog(LOG_WARNING, "main: protocol->write");
        daemon_exit_cleanly(-1);
      }
      continue;
    }

    
    /*
     * If local authentcation is uesed then now is the time to
     * Become someone else
     */
    if(opt.authenticate_in){
      if(geteuid()){
        syslog(
          LOG_INFO, 
          "Warning: not invoked as root, not setting uid and gid"
        );
      }
      else if(daemon_setid(opt.username, opt.group)){
        syslog(LOG_WARNING, "Error setting group and userid, Exiting");
        daemon_exit_cleanly(-1);
      }
    }


    /* Talk to the real pop server for the client*/
    if((server=tcp_socket_client_open(servername, port))<0){
      syslog(
        LOG_INFO, 
        "Could not connect to server, either server is "
        "unavailable or there is an invalid entry in %s for user %s",
        opt.mapname,
        pw->pw_name
      );
      sleep(PERDITION_ERR_SLEEP);
      if((*(protocol->write))(
        client_out,
	tag,
	protocol->type[PROTOCOL_ERR], 
	"Could not connect to server"
      )<0){
        syslog(LOG_WARNING, "main: protocol->write");
        daemon_exit_cleanly(-1);
      }
      continue;
    }

    /*We don't need this anymore*/
    server_port_destroy(server_port);
  
    /*Authenticate the user with the pop server*/
    status= (*(protocol->out_authenticate))(server, server, pw, protocol);
    if(status==0){
      sleep(PERDITION_ERR_SLEEP);
      syslog(LOG_INFO, "Fail reauthentication for user %s", pw->pw_name);
      free(pw->pw_name);
      free(pw->pw_passwd);
      free(pw);
      quit(server, server, protocol);
      if(close(server)){
        syslog(LOG_WARNING, "main: close(server) 2");
	daemon_exit_cleanly(-1);
      }
      if(protocol->write(
        client_out, 
        NULL, 
        protocol->type[PROTOCOL_NO], 
        "Re-Authentecation Failure"
      )<0){
        syslog(LOG_WARNING, "main: protocol->write");
        daemon_exit_cleanly(-1);
      }
      continue;
    }
    else if(status<0){
      syslog(LOG_WARNING, "main: protocol->out_authenticate %d", status);
      daemon_exit_cleanly(-1);
    }

    /*If we get this far, dance for joy with lmf*/
    if(protocol->write(
      client_out, 
      tag, 
      protocol->type[PROTOCOL_OK], 
      "You are so in"
    )<0){
      syslog(LOG_WARNING, "main: protocol->write");
      daemon_exit_cleanly(-1);
    }

    break;
  }

  /*We need a buffer for reads and writes to the server*/
  if((buffer=(unsigned char *)malloc(BUFFER_SIZE*sizeof(unsigned char)))==NULL){
    syslog(LOG_WARNING, "main: malloc: %d, Exiting", strerror(errno));
    daemon_exit_cleanly(-1);
  }

  /*Let the client talk to the real server*/
  if(tcp_socket_pipe(server,server,client_in,client_out,buffer,BUFFER_SIZE)<0){
    syslog(LOG_WARNING, "main: tcp_socket_pipe, Exiting (would anyway :)");
    daemon_exit_cleanly(-1);
  }

  /*Time to leave*/
  syslog(LOG_INFO, "Closing session for user %s", pw->pw_name);
  daemon_exit_cleanly(0);

  /*Here so compilers won't barf*/
  return(0);
}
